import copy

import pytest

import yapytf._containers as cnt


@pytest.fixture
def cmp(inp, out):
    if out is Ellipsis:
        out = copy.deepcopy(inp)

    def f(actual_out):
        assert actual_out == out

    return f


@pytest.mark.parametrize(
    'inp,out',
    [
        ({}, {}),
        ({}, ...)
    ]
)
def test_null(inp, cmp):
    cmp(inp)


# TODO
# test "generic was not parametrized" assertion


def test_DirectMutableDictAccessor_subclass():
    class D(cnt.DirectMutableDictAccessor[str, "cnt.DirectMutableDictAccessor[int, int]"]):
        _auto_create = True

    d = {}
    x = D(d)
    x["a"][1] = 2

    assert d == {"a": {1: 2}}


@pytest.mark.parametrize(
    'inp,out,tp,value',
    [
        ([], [1], int, 1),
        ([], [""], str, ""),
        ([], [[]], list, []),
    ]
)
def test_direct_append(inp, cmp, tp, value):
    x = cnt.DirectMutableListAccessor[tp](inp)
    x.append(value)
    cmp(inp)


@pytest.mark.parametrize(
    'inp,tp,value',
    [
        ([], str, 1),
        ([], int, ""),
    ]
)
def test_direct_append_wrong_type(inp, tp, value):
    with pytest.raises(TypeError):
        x = cnt.DirectMutableListAccessor[tp](inp)
        x.append(value)


@pytest.mark.parametrize(
    'inp,tkey,key,tvalue,value',
    [
        ({}, str, 1, str, ""),
        ({}, int, "", str, ""),
        ({}, int, 1, str, 1),
        ({}, int, 1, int, ""),
        ({}, "int", 1, int, ""),
    ]
)
def test_direct_set_wrong_type(inp, tkey, key, tvalue, value):
    with pytest.raises(TypeError):
        x = cnt.DirectMutableListAccessor[tkey, tvalue](inp)
        x[key] = value


def test_mutable_namespace():
    class O1(cnt.DirectMutableObjectAccessor):
        a: cnt.DictItemMutableListAccessor[int]

    class O2(cnt.DirectMutableObjectAccessor):
        b: cnt.DictItemMutableDictAccessor[int, str]

    class Ns(cnt.MutableNamespace):
        _paths = {
            'v': ('x', 'y')
        }
        u: O1
        v: O2

    d = {}
    ns = Ns(d)
    ns.u.a.append(1)
    ns.v.b[1] = "Z"
    assert d == {"a": [1], "x": {"y": {"b": {1: "Z"}}}}



import json
import pathlib
import shutil
from typing import Any, Dict

from . import _pcache, _tfrun


def get(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
    terraform_version: str,
    provider_name: str,
    provider_path: pathlib.Path,
    provider_version: str,
) -> Dict[str, Any]:
    key = "tfschema-{}-{}-{}".format(terraform_version, provider_name, provider_version)

    FNAME = "schema.json"

    def produce(dir_path: pathlib.Path) -> None:
        this_work_dir = work_dir.joinpath("tfschema-{}".format(provider_name))
        this_work_dir.mkdir()

        with this_work_dir.joinpath("main.tf.json").open("w") as f:
            json.dump({"provider": [{provider_name: {}}]}, f)

        _tfrun.tf_init(
            work_dir=this_work_dir,
            terraform_path=terraform_path,
            providers_paths=[provider_path]
        )

        output = _tfrun.tf_run_non_interactive(
            work_dir=this_work_dir,
            terraform_path=terraform_path,
            args=["providers", "schema", "-json"]
        )
        dir_path.joinpath(FNAME).write_bytes(output)

    cache_dir = _pcache.get(key, produce)

    try:
        with cache_dir.joinpath(FNAME).open() as f:
            schema = json.load(f)

        if not isinstance(schema, dict):
            raise RuntimeError(
                'terraform schema for "{}" provider is not a json object'.format(
                    provider_name
                )
            )
    except Exception:
        shutil.rmtree(cache_dir)
        raise

    return schema



import textwrap as _textwrap
import typing as _typing


def dedent(s: str) -> str:
    return _textwrap.dedent(s).lstrip()


def format(
    template_str: str,
    *args: _typing.Any,
    **kw: _typing.Any
) -> str:
    return dedent(template_str).format(*args, **kw)



import json
import typing as _tp

from . import JsonType
from . import _containers as _cnt


def dict_path_rw(
    d: JsonType,
    path: _tp.Iterable[str]
) -> _tp.Dict[str, _tp.Any]:
    for i in path:
        d = d.setdefault(i, {})

    return d


def dict_path_ro(
    d: JsonType,
    path: _tp.Iterable[str]
) -> _tp.Dict[str, _tp.Any]:
    for i in path:
        d = d.get(i, {})

    return d


class Locals(_tp.MutableMapping[str, _tp.Any]):
    def __init__(self, data: _tp.Dict[str, _tp.Any]) -> None:
        self._data = dict_path_rw(data, ["tf", "locals"])

    def __delitem__(self, key: str) -> None:
        self._data.pop(key)

    def __getitem__(self, key: str) -> _tp.Any:
        return self._data[key]

    def __iter__(self) -> _tp.Iterator[str]:
        return self._data.__iter__()

    def __len__(self) -> int:
        return len(self._data)

    def __setitem__(self, key: str, value: _tp.Any) -> None:
        # note: this does not preclude further data manipulations from
        # making the value non-json-serializable, yet this is
        # a test to catch most straighforward mistakes
        json.dumps(value)
        self._data[key] = value


@_tp.runtime_checkable
class Extras(_tp.Protocol):
    # regretfully, typing is not yet powerful enough to express
    # requirement that the number of *args must match the number
    # of args in Callable signature. Also, we want a true "json serializable"
    # type instead of Any

    def func(self, func: _tp.Callable[..., _tp.Any], *args: _tp.Any) -> str:
        ...


class _DictWithDefault(_cnt.PlainDict[_cnt.Tkey, _cnt.Tval]):
    @property
    def default(self) -> _cnt.Tval:
        return self[_tp.cast(_cnt.Tkey, "")]

    @default.setter
    def default(self, val: _cnt.Tval) -> None:
        self[_tp.cast(_cnt.Tkey, "")] = val

    @default.deleter
    def default(self) -> None:
        del self[_tp.cast(_cnt.Tkey, "")]


class _ConstDictWithX(_cnt.ConstDict[_cnt.Tkey, _cnt.Tval]):
    @property
    def x(self) -> _cnt.Tval:
        return self[_tp.cast(_cnt.Tkey, None)]


class _SolitaryZerothAsXMixin(_tp.Sequence[_cnt.Tval]):
    @property
    def x(self) -> _cnt.Tval:
        assert len(self) == 1
        return self[0]



import dataclasses
import functools
import http.server
import importlib
import inspect
import json
import logging
import pathlib
import runpy
import shutil
import sys
import tempfile
import threading
from typing import (Any, Callable, Dict, Generator, Iterable, List, Mapping,
                    Optional, Tuple, Type, cast)
import urllib
import uuid

import click
import click_log
import yaml

import jsonschema
import mypy.api
import toposort
from implements import implements

from . import (Configurator, JsonType, StateBackendConfig, _genbase,
               _generator, _hashigetter, _tfrun, _tfschema, _tfstate)

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

VERSIONS_SCHEMA = {
    "type": "object",
    "properties":
    {
        "terraform": {"type": "string"},
        "providers":
        {
            "type": "object",
            "additionalProperties": {"type": "string"},
        },
    },
    "required": ["terraform"],
    "additionalProperties": False
}


# Per https://stackoverflow.com/questions/17211078/how-to-temporarily-modify-sys-path-in-python
class AddSysPath:
    def __init__(self, path: str):
        self.path = path

    def __enter__(self) -> None:
        sys.path.append(self.path)

    def __exit__(self, *args: Any) -> None:
        try:
            sys.path.remove(self.path)
        except ValueError:
            pass


def wipe_dir(d: pathlib.Path) -> None:
    for i in d.iterdir():
        if i.is_dir():
            shutil.rmtree(i)
        else:
            i.unlink()


class Extras(_genbase.Extras):
    def __init__(
        self,
        model: Any,
        http_server_address: Tuple[str, int] = ("", 0),
    ) -> None:
        self.model = model

        self._funcs: List[Callable[..., Any]] = []
        self._http_server_address = f"{http_server_address[0]}:{http_server_address[1]}"
        self._http_server_secret = str(uuid.uuid4())

    def _http_get(self, path: str) -> Optional[str]:
        parsed = urllib.parse.urlparse(path)

        if parsed.path != "/":
            return None

        query = urllib.parse.parse_qs(parsed.query)

        if query.get("s") != [self._http_server_secret]:
            return None

        f = self._funcs[int(query.get("f", [])[0])]
        a = json.loads(query.get("a", [])[0])
        r = f(*a)

        return json.dumps(r)

    def func(self, func: Callable[..., Any], *args: str) -> str:
        func_n = len(self._funcs)
        self._funcs.append(func)
        id = f"_yapytf{func_n:05}"
        v1 = self.model.tf.v1
        args_s = ", ".join(args)
        v1.d.http.X[id].url = (
            f"http://{self._http_server_address}"
            f"/?s={self._http_server_secret}&f={func_n}&a=${{urlencode(jsonencode([{args_s}]))}}"
        )
        v1.l[id] = f"${{jsondecode(data.http.{id}.body)}}"
        return f"${{local.{id}}}"


class Model:
    _versions: Dict[str, Any]
    _yapytffile_path: pathlib.Path
    _terraform_path: Optional[pathlib.Path]
    _providers_paths: Optional[Mapping[str, pathlib.Path]]
    _providers_schemas: Optional[Mapping[str, Dict[str, Any]]]
    _work_dir: pathlib.Path
    _resource_type_to_provider: Optional[Dict[str, Dict[str, str]]]
    _resources_schemas_versions: Optional[Dict[Tuple[str, str], int]]
    _configurator_classes: List[Type[Configurator]]

    def __init__(
        self,
        *,
        path: pathlib.Path,
        model_classes: Iterable[str],
        work_dir: pathlib.Path,
    ):
        self._yapytffile_path = path
        self._work_dir = work_dir

        with AddSysPath(str(path.parent)):
            logger.debug("running %s, sys.path=%s", path, sys.path)
            py = runpy.run_path(str(path))

        if "yapytfgen" in sys.modules:
            raise click.ClickException(
                f'"{path}" has imported "yapytfgen" module. Please wrap the import with "if typing.TYPE_CHECKING".'
            )

        def get_class(name: str) -> Configurator:
            class_ = py.get(name)
            if not inspect.isclass(class_):
                raise click.ClickException(f'"{name}" is not defined or is not a class')

            return cast(Configurator, implements(Configurator)(class_))

        classes = {name: get_class(name) for name in model_classes}
        requires = {name: class_.requires() for name, class_ in classes.items()}
        unsatisfied = [
            f"{name}->{j}"
            for name, i in requires.items()
            for j in i
            if j not in model_classes
        ]
        if unsatisfied:
            raise click.ClickException("Unsatisfied model class requirements: " + ", ".join(unsatisfied))

        order = toposort.toposort_flatten(requires)
        logger.debug("sorted model classes: %s", order)

        self._configurator_classes = [classes[i] for i in order]

        self._versions = dict(
            providers={
                "http": "1.2.0",
            },
        )
        for class_ in self._configurator_classes:
            class_.versions(self.versions)

        try:
            jsonschema.validate(self.versions, VERSIONS_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise click.ClickException(e)

        self._terraform_path = None
        self._providers_paths = None
        self._providers_schemas = None
        self._resource_type_to_provider = None
        self._resources_schemas_versions = None

    @property
    def versions(self) -> Dict[str, Any]:
        return self._versions

    @property
    def terraform_version(self) -> str:
        return cast(str, self.versions["terraform"])

    @property
    def yapytffile_path(self) -> pathlib.Path:
        return self._yapytffile_path

    @property
    def terraform_path(self) -> pathlib.Path:
        if self._terraform_path is None:
            self._terraform_paths = _hashigetter.get_terraform(self.terraform_version)

        return self._terraform_paths

    @property
    def providers_versions(self) -> Mapping[str, str]:
        return self.versions.get("providers", {})

    @property
    def providers_paths(self) -> Mapping[str, pathlib.Path]:
        if self._providers_paths is None:
            self._providers_paths = dict(
                (name, _hashigetter.get_terraform_provider(name, ver))
                for name, ver in self.providers_versions.items()
            )

        return self._providers_paths

    @property
    def providers_schemas(self) -> Mapping[str, Dict[str, Any]]:
        if self._providers_schemas is None:
            self._providers_schemas = {
                provider_name: _tfschema.get(
                    work_dir=self._work_dir,
                    terraform_path=self.terraform_path,
                    terraform_version=self.terraform_version,
                    provider_name=provider_name,
                    provider_version=provider_version,
                    provider_path=self.providers_paths[provider_name]
                )
                for provider_name, provider_version in self.providers_versions.items()
            }

        return self._providers_schemas

    @property
    def resource_type_to_provider(self) -> Dict[str, Dict[str, str]]:
        if self._resource_type_to_provider is None:
            self._resource_type_to_provider = {
                kind_key: {
                    resource_type: provider_name
                    for provider_name, provider_schema in self.providers_schemas.items()
                    for resource_type in provider_schema["provider_schemas"][provider_name].get(f"{kind}_schemas", {})
                }
                for kind, kind_key in _generator.KIND_TO_KEY.items()
            }

        return self._resource_type_to_provider

    @property
    def resources_schemas_versions(self) -> Dict[Tuple[str, str], int]:
        if self._resources_schemas_versions is None:
            self._resources_schemas_versions = {
                (kind_key, resource_type): resource_schema["version"]
                for kind, kind_key in _generator.STATE_KIND_TO_KEY.items()
                for provider_name, provider_schema in self.providers_schemas.items()
                for resource_type, resource_schema
                in provider_schema["provider_schemas"][provider_name].get(f"{kind}_schemas", {}).items()
            }

        return self._resources_schemas_versions

    def gen_yapytfgen(self, module_dir: pathlib.Path) -> None:
        logger.debug("terraform_path: %s", self.terraform_path)
        logger.debug("providers_paths: %s", self.providers_paths)

        providers_py_paths: Dict[str, pathlib.Path] = {}

        for provider_name, provider_version in self.providers_versions.items():
            providers_py_paths[provider_name] = _generator.gen_provider_py(
                work_dir=self._work_dir,
                terraform_version=self.terraform_version,
                provider_name=provider_name,
                provider_version=provider_version,
                provider_schema=self.providers_schemas[provider_name]
            )

        _generator.gen_yapytfgen(
            module_dir=module_dir,
            providers_paths=providers_py_paths,
        )

    @dataclasses.dataclass
    class PreparedStep:
        path: pathlib.Path
        http_get_handler: Callable[[str], Optional[str]]

    @dataclasses.dataclass
    class PreparedSteps:
        configurators: List[Configurator]
        steps: List["Model.PreparedStep"]

    def prepare_steps(
        self,
        *,
        model_params: Dict[str, Any],
        destroy: bool = False,
        http_server_address: Tuple[str, int] = ("", 0),
    ) -> "Model.PreparedSteps":
        schema: Dict[str, Any] = {
            "$schema": "http://json-schema.org/schema#",
            "type": "object",
            "properties": {},
            "required": [],
            "additionalProperties": False
        }

        for class_ in self._configurator_classes:
            class_.schema(schema)

        try:
            jsonschema.validate(model_params, schema)
        except jsonschema.exceptions.ValidationError as e:
            raise click.ClickException(e)

        configurators = [class_(model_params) for class_ in self._configurator_classes]

        state_backend_cfg = StateBackendConfig()

        for i in configurators:
            i.state_backend_cfg(state_backend_cfg)

        steps: List[Tuple[str, Any]]

        if destroy:
            steps = [("destroy", None)]
        else:
            steps = [("create1", None)]

        module_dir = self._work_dir.joinpath("yapytfgen")
        module_dir.mkdir()

        self.gen_yapytfgen(module_dir)

        with AddSysPath(str(self._work_dir)):
            self.yapytfgen_module = importlib.import_module("yapytfgen")

        yapytfgen_model_class = getattr(self.yapytfgen_module, "model")
        yapytfgen_providers_model_class = getattr(self.yapytfgen_module, "providers_model")

        prepared_steps = []
        for step_name, step_data in steps:
            step_dir = self._work_dir.joinpath(f"step.{step_name}")
            step_dir.mkdir()

            d: Dict[str, Any] = {
                "provider": {
                    provider_name: {'': {}}
                    for provider_name in self.providers_versions
                },
                "tf": {},
            }

            yapytfgen_providers_model = yapytfgen_providers_model_class(d)
            for i in configurators:
                i.populate_providers(model=yapytfgen_providers_model)

            if not destroy:
                yapytfgen_model = yapytfgen_model_class(d)
                extras = d["extras"] = Extras(yapytfgen_model, http_server_address)
                http_get_handler = extras._http_get

                for i in configurators:
                    i.populate(model=yapytfgen_model, step_data=step_data)
            else:
                http_get_handler = lambda path: None

            assert "provider" not in d["tf"]
            step_tf_data = {
                "terraform": {
                    "backend": {
                        state_backend_cfg.name: state_backend_cfg.vars
                    }
                },
                "provider": [
                    {
                        provider_type:
                        {
                            **provider_data,
                            **({"alias": provider_alias} if provider_alias else {})
                        }
                    }
                    for provider_type, provider_aliases in d["provider"].items()
                    for provider_alias, provider_data in provider_aliases.items()
                ],
            }

            def drop_empty(d: Dict[Any, Any]) -> Generator[Any, None, None]:
                return ((k, v) for k, v in d.items() if v)

            step_tf_data.update(drop_empty({k: dict(drop_empty(v)) for k, v in d["tf"].items()}))

            # expand "provider" properties
            for kind in ["data", "resource"]:
                resource_type_to_provider = self.resource_type_to_provider[kind]
                for resource_type, resources in step_tf_data.get(kind, {}).items():
                    for resource_name, resource_data in resources.items():
                        provider_alias = resource_data.get("provider")
                        if provider_alias is not None:
                            provider_type = resource_type_to_provider[resource_type]
                            assert provider_alias in d["provider"][provider_type]
                            resource_data["provider"] = f"{provider_type}.{provider_alias}"

            with step_dir.joinpath("main.tf.json").open("w") as f:
                json.dump(step_tf_data, f, indent=4, sort_keys=True)

            with step_dir.joinpath("debug.tf.yaml").open("w") as f:
                yaml.dump(step_tf_data, default_flow_style=False, stream=f)

            _tfrun.tf_init(
                work_dir=step_dir,
                terraform_path=self.terraform_path,
                providers_paths=self.providers_paths.values()
            )

            _tfrun.tf_validate(
                work_dir=step_dir,
                terraform_path=self.terraform_path,
            )

            prepared_steps.append(Model.PreparedStep(step_dir, http_get_handler))

        return Model.PreparedSteps(configurators, prepared_steps)


class PathType(click.Path):
    def coerce_path_result(self, rv) -> pathlib.Path:  # type: ignore
        return pathlib.Path(super().coerce_path_result(rv))


def click_wrapper(wrapper: Callable[..., None], wrapped: Callable[..., None]) -> Callable[..., None]:
    wrapped_params = getattr(wrapped, "__click_params__", [])
    wrapper_params = getattr(wrapper, "__click_params__", [])
    result = functools.update_wrapper(wrapper, wrapped)
    result.__click_params__ = wrapped_params + wrapper_params  # type: ignore
    return result


def base_command(func: Callable[..., None]) -> Callable[..., None]:
    @click_log.simple_verbosity_option(logger)  # type: ignore
    def wrapper(**opts: Any) -> None:
        func(**opts)

    return click_wrapper(wrapper, func)


def model_class_command(func: Callable[..., None]) -> Callable[..., None]:
    @base_command
    @click.option(
        "--dir",
        "-C",
        type=PathType(dir_okay=True, exists=True),
        default=".",
        help="Directory containing the model file, instead of the current working directory.",
    )
    @click.option(
        "--file",
        "-f",
        metavar="NAME",
        default="Yapytffile.py",
        help="Name of a model file to use.",
        show_default=True,
    )
    @click.option(
        "--classes",
        metavar="C1[,C2...]",
        default="Default",
        help="Comma separated list of names of a model classes to use.",
        show_default=True,
    )
    @click.option("--keep-work", is_flag=True, help="Keep working directory.")
    def wrapper(**opts: Any) -> None:
        model_path = opts["dir"].joinpath(opts["file"])
        if not model_path.exists():
            raise click.FileError(str(model_path), hint="no such file")

        work_dir = pathlib.Path(tempfile.mkdtemp(prefix="yapytf."))

        try:
            model = Model(
                path=model_path,
                model_classes=opts["classes"].split(","),
                work_dir=work_dir,
            )
            func(model, **opts)
        finally:
            if opts["keep_work"]:
                click.echo('Keeping working directory "{}"'.format(work_dir), err=True)
            else:
                shutil.rmtree(work_dir)

    return click_wrapper(wrapper, func)


def model_instance_command(func: Callable[..., None]) -> Callable[..., None]:
    @model_class_command
    @click.option(
        "--params",
        type=PathType(file_okay=True, exists=True),
        help='json file containing model parameters passed to constructors of the model classes.',
    )
    def wrapper(model: Model, **opts: Any) -> None:
        if opts["params"]:  # types: pathlib.Path
            with opts["params"].open() as f:
                try:
                    model_params = json.load(f)
                except json.JSONDecodeError as e:
                    raise click.ClickException("{}: {}".format(opts["params"], e))

            if not isinstance(model_params, dict):
                raise click.ClickException(
                    "{}: expected a json object".format(opts["params"])
                )
        else:
            model_params = {}

        func(model, model_params, **opts)

    return click_wrapper(wrapper, func)


@click.group()
@click.version_option()
def main(**opts: Any) -> None:
    """
    Yet Another Python Terraform Wrapper
    """


@main.command(hidden=True)
@click.argument("product")
@click.argument("ver")
def get(**opts: Any) -> None:
    """
    Download hashicorp product.
    """
    d = _hashigetter.get_hashi_product(opts["product"], opts["ver"])
    logger.info("got: %s", d)


@main.command()
@model_class_command
def dev(model: Model, **opts: Any) -> None:
    """
    Set up development environment.

    Sets up development environment in directory containing Yapytffile.py
    by (re-)generating "yapytfgen" module and symlinking "yapytf" module.
    This should allow auto-completion (typically based on "jedi")
    and type checkers such a mypy to "just work".
    Warning, existing "yapytfgen" directory will be completely wiped out.
    As a safeguard, a ".yapytfgen" marker file must exist in that directory.
    Any existing symlink named "yapytf" will be overwritten.
    """

    dest_dir: pathlib.Path = model.yapytffile_path.parent

    # yapytf symlink

    yapytf_target = pathlib.Path(__file__).parent
    yapytf_link = dest_dir.joinpath("yapytf")

    if yapytf_link.exists():
        if not yapytf_link.is_symlink():
            raise click.ClickException(
                f"Cowardly refusing to overwrite existing \"{yapytf_link}\", "
                + "which is not a symlink"
            )

        yapytf_link.unlink()

    yapytf_link.symlink_to(yapytf_target, target_is_directory=True)

    # yapytfgen module

    yapytfgen_dir: pathlib.Path = dest_dir.joinpath("yapytfgen")
    marker_file: pathlib.Path = yapytfgen_dir.joinpath(".yapytfgen")
    if yapytfgen_dir.exists():
        if not marker_file.exists():
            raise click.ClickException(
                f"Cowardly refusing to wipe existing \"{yapytfgen_dir}\", "
                + "which does not contain \".yapytfgen\" marker file"
            )

        shutil.rmtree(yapytfgen_dir)

    yapytfgen_dir.mkdir()
    marker_file.touch()
    yapytfgen_dir.joinpath(".gitignore").write_text("*\n")

    model.gen_yapytfgen(module_dir=yapytfgen_dir)


@main.command()
@model_instance_command
def lint(model: Model, model_params: JsonType, **opts: Any) -> None:
    """
    Validate model file.
    """

    # TODO chose one of the following:
    # 1. automatically run "dev" command
    # 2. fail early and advice to run "dev" if either "yapytfgen" or "yapytf" symlinks is missing
    # 3. refactor the flow to have lint executed after workdir has been prepared and
    #    MYPYPATH env var is set accoringly.

    mypy_out, mypy_err, mypy_rc = mypy.api.run([
        str(model.yapytffile_path),
        "--cache-dir",
        str(model.yapytffile_path.with_name(".mypy_cache")),
        # "--no-incremental",
        "--strict",
        "--allow-redefinition",
        "--follow-imports=silent",
        "--implicit-reexport",
    ])
    if mypy_rc:
        sys.stdout.write(mypy_out)
        sys.stderr.write(mypy_err)

    model.prepare_steps(model_params=model_params)

    if mypy_rc:
        sys.exit(1)


@main.command()
@model_instance_command
@click.option(
    "--out",
    type=PathType(dir_okay=True),
    help="Directory into which to store model outputs. Warning: will completely wipe the directory! "
    "Note that outputs will only be written in case of a successful execution."
)
def apply(model: Model, model_params: JsonType, **opts: Any) -> None:
    """
    Build or change infrastructure.
    """

    http_get_handlers = []

    class HTTPRequestHandler(http.server.BaseHTTPRequestHandler):
        def log_message(self, format: str, *args: Any) -> None:
            logger.debug("http server: {}".format(format % args))

        def do_GET(self) -> None:
            if not http_get_handlers:
                self.send_response(404)
            else:
                result = http_get_handlers[-1](self.path)
                self.send_response(404 if result is None else 200)
                self.send_header("Content-type", "text/text")
                self.end_headers()
                if result is not None:
                    self.wfile.write(result.encode())

    http_server = http.server.ThreadingHTTPServer(("localhost", 0), HTTPRequestHandler)

    def http_server_func() -> None:
        http_server.serve_forever(poll_interval=0.1)

    http_server_thread = threading.Thread(target=http_server_func)
    http_server_thread.start()

    try:
        steps = model.prepare_steps(model_params=model_params, http_server_address=http_server.server_address)
        for step in steps.steps:
            http_get_handlers.append(step.http_get_handler)
            if _tfrun.tf_apply(work_dir=step.path, terraform_path=model.terraform_path):
                st = _tfrun.tf_get_state(work_dir=step.path, terraform_path=model.terraform_path)
                st = _tfstate.get_resources_attrs(st, model.resources_schemas_versions)

                yapytfgen_state_class = getattr(model.yapytfgen_module, "state")
                state = yapytfgen_state_class(st)

                out_dir: Optional[pathlib.Path] = opts["out"]
                if out_dir is not None:
                    marker_file: pathlib.Path = out_dir.joinpath(".yapytf_out")

                    if out_dir.exists():
                        if not marker_file.exists():
                            raise click.ClickException(
                                f"Cowardly refusing to wipe existing \"{out_dir}\", "
                                + "which does not contain \".yapytf_out\" marker file"
                            )
                        wipe_dir(out_dir)
                    else:
                        out_dir.mkdir()

                    marker_file.touch()

                    for i in steps.configurators:
                        i.output(state=state, dest=out_dir)
            else:
                sys.exit(1)
    finally:
        http_server.shutdown()
        http_server_thread.join()


@main.command()
@model_instance_command
def destroy(model: Model, model_params: JsonType, **opts: Any) -> None:
    """
    Destroy Terraform-managed infrastructure.
    """

    steps = model.prepare_steps(model_params=model_params, destroy=True)
    assert len(steps.steps) == 1
    _tfrun.tf_destroy(work_dir=steps.steps[0].path, terraform_path=model.terraform_path)


@main.command()
@model_instance_command
@click.argument(
    "args",
    nargs=-1
)
def rawtf(model: Model, model_params: JsonType, **opts: Any) -> None:
    """
    Execute raw terraform commands.
    """
    # TODO



import json
import pathlib
import shlex
import subprocess
import sys
from typing import cast, Any, Dict, Iterable

import click

TF_BIN = "terraform"


def tf_run(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
    capture_output: bool,
    args: Iterable[str],
) -> subprocess.CompletedProcess:
    return subprocess.run(
        [str(terraform_path.joinpath(TF_BIN))] + list(args),
        cwd=work_dir,
        capture_output=capture_output
    )


def tf_raise(
    returncode: int,
    args: Iterable[str],
) -> None:
    if returncode:
        raise click.ClickException("\"terraform{}\" failed with return code {}".format(
            "".join(f" {shlex.quote(i)}" for i in args),
            returncode
        ))


def tf_run_non_interactive(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
    args: Iterable[str],
) -> bytes:
    cp = tf_run(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=args,
        capture_output=True
    )

    if cp.returncode:
        WIDTH = 80
        sys.stderr.writelines([" terraform output ".center(WIDTH, "="), "\n"])
        sys.stderr.buffer.write(cp.stdout)
        sys.stderr.buffer.write(cp.stderr)
        sys.stderr.writelines(["".center(WIDTH, "="), "\n"])

        tf_raise(cp.returncode, args)

    return cast(bytes, cp.stdout)


def tf_run_interactive(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
    args: Iterable[str],
) -> int:
    cp = tf_run(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=args,
        capture_output=False
    )
    return cp.returncode


def tf_init(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
    providers_paths: Iterable[pathlib.Path],
) -> None:
    tf_run_non_interactive(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=["init"] + [i for provider_path in providers_paths for i in ["-plugin-dir", str(provider_path)]]
    )


def tf_validate(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
) -> None:
    tf_run_non_interactive(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=["validate"]
    )


def tf_get_state(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
) -> Dict[str, Any]:
    output = tf_run_non_interactive(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=["state", "pull"]
    )
    json_output = json.loads(output)
    if not isinstance(json_output, dict):
        raise RuntimeError('Output of "terraform state pull" is not a json object')

    return json_output


def tf_apply(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
) -> bool:
    returncode = tf_run_interactive(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=[
            "apply",
            "-input=false"
        ]
    )
    return returncode == 0


def tf_destroy(
    *,
    work_dir: pathlib.Path,
    terraform_path: pathlib.Path,
) -> None:
    args = [
        "destroy",
        "-input=false"
    ]
    returncode = tf_run_interactive(
        work_dir=work_dir,
        terraform_path=terraform_path,
        args=args
    )
    tf_raise(returncode, args)



from ._public import *  # noqa: F401, F403

try:
    __version__ = __import__('pkg_resources').get_distribution(__name__).version
except:  # noqa # pragma: no cover
    pass

APPNAME = __name__
APPAUTHOR = "Kyrylo Shpytsya"



import errno
import contextlib
import os
import pathlib
import shutil
import urllib.parse
from typing import Callable

import appdirs

from . import APPAUTHOR, APPNAME

CACHE_DIR = pathlib.Path(appdirs.user_cache_dir(appname=APPNAME, appauthor=APPAUTHOR))


def get(key: str, produce_cb: Callable[[pathlib.Path], None]) -> pathlib.Path:
    dir_name = urllib.parse.quote(key, safe="")
    dir_path = CACHE_DIR.joinpath(dir_name)

    CACHE_DIR.mkdir(exist_ok=True, parents=True)

    if not dir_path.exists():
        tmp_dir_name = "{}.tmp{}".format(dir_name, os.getpid())
        tmp_dir_path = CACHE_DIR.joinpath(tmp_dir_name)

        with contextlib.ExitStack() as es:
            tmp_dir_path.mkdir()

            @es.callback
            def cleanup_tmp_dir():
                shutil.rmtree(tmp_dir_path)

            produce_cb(tmp_dir_path)

            # in case of race when othen process has managed to produce the same
            # item before this process, let the former win
            if not dir_path.exists():
                renamed = True

                try:
                    tmp_dir_path.rename(dir_path)
                except OSError as e:
                    if e.errno == errno.ENOTEMPTY:
                        renamed = False
                    else:
                        raise

                if renamed:
                    es.pop_all()

    return dir_path



import copy
import typing as tp

import pytest

import yapytf._containers2 as cnt


@pytest.yield_fixture
def cmp(inp, out):
    if out is Ellipsis:
        out = copy.deepcopy(inp)

    yield

    assert inp == out


@pytest.mark.parametrize(
    'inp,out',
    [
        ({}, {}),
        ({}, ...)
    ]
)
def test_nop(cmp):
    pass


@pytest.mark.parametrize(
    'bad_container',
    [
        [],
        set(),
        1,
        ""
    ]
)
def test_Record_bad_container(bad_container):
    with pytest.raises(TypeError):
        cnt.Record(bad_container)


@pytest.mark.parametrize(
    'inp,out',
    [
        ({"f1": 1, "f2_": "2", "f3": "a", "f4": {"f4a": {"f4b": 4}}}, ...)
    ]
)
def test_ConstRecordField(inp, cmp):
    class Rec2(cnt.Record):
        f1 = cnt.ConstRecordField[str]()

    with pytest.raises(RuntimeError):
        class BadRecord(cnt.Record):
            # missing generic arg
            f1 = cnt.ConstRecordField()

    class IntField(cnt.ConstRecordField[int]):
        pass

    class Rec1(cnt.Record):
        f1 = IntField()

    class Rec(cnt.Record):
        f1 = IntField()
        f2 = cnt.ConstRecordField[str](name="f2_")
        f3 = IntField()
        f4 = IntField(path=["f4", "f4a"], name="f4b")
        f5 = IntField()

    r = Rec(inp)

    assert r.f1 == 1
    assert r.f2 == "2"
    assert r.f4 == 4

    with pytest.raises(NotImplementedError):
        r.f1 = 1

    with pytest.raises(NotImplementedError):
        del r.f1

    with pytest.raises(TypeError):
        r.f3

    with pytest.raises(AttributeError):
        r.f5

    r2 = Rec(inp, path=["nonexistent"])

    with pytest.raises(KeyError):
        r2.f1


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {"f1": 1, "f3": []},
            {"f1": 2, "f2": "a", "x": {"f1": 3}})
    ]
)
def test_PlainRecordField(inp, cmp):
    class Rec(cnt.Record):
        f1 = cnt.PlainRecordField[int]()
        f2 = cnt.PlainRecordField[str]()
        f3 = cnt.PlainRecordField[str]()

    r = Rec(inp)

    assert r.f1 == 1
    with pytest.raises(AttributeError):
        r.f2

    r2 = Rec(inp)

    r.f1 = 2
    r.f2 = "a"
    del r.f3

    assert r2.f1 == 2
    assert r2.f2 == "a"
    with pytest.raises(AttributeError):
        r2.f3

    r3 = Rec(inp, path=["x"])
    r3.f1 = 3

    with pytest.raises(TypeError):
        r3.f1 = ""


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "int2str": {1: "A", 2: 3}
            },
            ...
        )
    ]
)
def test_ConstRecordField_ConstDict(inp, cmp):
    class R(cnt.Record):
        int2str = cnt.ConstRecordField[cnt.ConstDict[int, str]]()

    r = R(inp)

    assert len(r.int2str) == 2
    assert set(r.int2str) == {1, 2}

    assert r.int2str[1] == "A"

    with pytest.raises(TypeError):
        r.int2str[2]

    with pytest.raises(KeyError):
        r.int2str[3]

    with pytest.raises(TypeError):
        r.int2str["x"]


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "int2str": {1: "A", "2": 3}
            },
            ...
        )
    ]
)
def test_ConstDict_bad_key(inp, cmp):
    class R(cnt.Record):
        int2str = cnt.ConstRecordField[cnt.ConstDict[int, str]]()

    r = R(inp)

    with pytest.raises(TypeError):
        list(r.int2str)


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "str2rec":
                {
                    "a":
                    {
                        "s": "S",
                        "i": 1,
                    }
                }
            },
            ...
        )
    ]
)
def test_ConstDict_Record(inp, cmp):
    class R2(cnt.Record):
        s = cnt.ConstRecordField[str]()
        i = cnt.ConstRecordField[int]()

    class R(cnt.Record):
        str2rec = cnt.ConstRecordField[cnt.ConstDict[str, R2]]()

    r = R(inp)
    assert len(r.str2rec) == 1
    list(r.str2rec)

    assert r.str2rec["a"].s == "S"
    assert r.str2rec["a"].i == 1

    with pytest.raises(NotImplementedError):
        del r.str2rec

    with pytest.raises(TypeError):
        del r.str2rec["a"]

    with pytest.raises(NotImplementedError):
        del r.str2rec["a"].s

    with pytest.raises(NotImplementedError):
        del r.str2rec["a"].i



@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1":
                {
                    "a":
                    {
                        1: b"A",
                        2: b"B"
                    }
                },
                "f2":
                {
                    "u": {10: {b"v": 1.5}}
                }
            },
            ...
        )
    ]
)
def test_ConstDict_nested_plain(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.ConstRecordField[cnt.ConstDict[str, cnt.ConstDict[int, bytes]]]()
        f2 = cnt.ConstRecordField[cnt.ConstDict[str, cnt.ConstDict[int, cnt.ConstDict[bytes, float]]]]()

    r = R(inp)
    assert len(r.f1) == 1
    list(r.f1)

    a = r.f1["a"]
    assert len(a) == 2
    assert set(a) == {1, 2}
    assert a[1] == b"A"
    assert a[2] == b"B"

    assert len(r.f2) == 1
    assert r.f2["u"][10][b"v"] == 1.5


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1": [1, 2, 3],
                "f2": ["a", "b"],
                "f3": ["a", 1, "b", 2],
            },
            ...
        )
    ]
)
def test_ConstDict_nested_plain(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.ConstRecordField[cnt.ConstList[int]]()
        f2 = cnt.ConstRecordField[cnt.ConstList[str]]()
        f3 = cnt.ConstRecordField[cnt.ConstList[str]]()

    r = R(inp)
    assert len(r.f1) == 3
    assert len(r.f2) == 2
    assert len(r.f3) == 4

    assert list(r.f1) == [1, 2, 3]
    assert list(r.f1[1:]) == [2, 3]
    assert list(r.f2) == ["a", "b"]
    assert list(r.f3[0::2]) == ["a", "b"]

    with pytest.raises(TypeError):
        r.f3[:]


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "a":
                {
                    "b":
                    {
                        "f1_":
                        {
                            "a": [1],
                            "ab": [1, 2],
                            "abc": [1, 2, 3]
                        }

                    }
                }
            },
            ...
        )
    ]
)
def test_ConstDict_ConstList(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.ConstRecordField[cnt.ConstDict[str, cnt.ConstList[int]]](name="f1_", path=["a", "b"])

    r = R(inp)
    assert len(r.f1) == 3

    assert set(r.f1) == {"a", "ab", "abc"}
    assert list(r.f1["a"]) == [1]
    assert list(r.f1["ab"]) == [1, 2]
    assert list(r.f1["abc"][::2]) == [1, 3]


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "fa":
                {
                    "f1": 1
                }
            },
            {
                "fa":
                {
                    "f1": 1
                },
                "fc":
                {
                    "f1": 2
                }
            }
        )
    ]
)
def test_RecordField_RecordField(inp, cmp):
    class R2(cnt.Record):
        f1 = cnt.PlainRecordField[int]()

    class R(cnt.Record):
        fa = cnt.PlainRecordField[R2]()
        fb = cnt.PlainRecordField[R2]()
        fc = cnt.PlainRecordField[R2]()

    r = R(inp)

    assert r.fa.f1 == 1
    r.fb

    with pytest.raises(AttributeError):
        r.fb.f1

    r.fc.f1 = 2


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1":
                {
                    "a": 1,
                    "b": 2
                },
                "f1bad":
                {
                    1: "a",
                    "2": "b"
                }
            },
            {
                "f1":
                {
                    "a": 1,
                    "c": 3
                },
                "f2":
                {
                    1: "a"
                }
            }
        )
    ]
)
def test_RecordField_PlainDict(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.PlainRecordField[cnt.PlainDict[str, int]]()
        f1bad = cnt.PlainRecordField[cnt.PlainDict[str, int]]()
        f2 = cnt.PlainRecordField[cnt.PlainDict[int, str]]()

    r = R(inp)

    assert len(r.f1) == 2
    assert sorted(r.f1.items()) == [("a", 1), ("b", 2)]
    del r.f1["b"]

    with pytest.raises(TypeError):
        list(r.f1bad)

    with pytest.raises(TypeError):
        r.f1bad["2"]

    del r.f1bad

    with pytest.raises(KeyError):
        del r.f1["b"]

    r.f1["c"] = 3

    with pytest.raises(AttributeError):
        len(r.f2)

    with pytest.raises(AttributeError):
        list(r.f2)

    r.f2[1] = "a"


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1": [1, 11]
            },
            {
                "f1": [10, 2, 3]
            }
        )
    ]
)
def test_List(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.RecordField[cnt.PlainList[int], tp.Iterable[int]]()

    r = R(inp)
    assert len(r.f1) == 2

    assert set(r.f1) == {1, 11}

    del r.f1[1]
    r.f1[0] = 10
    r.f1.append(2)
    r.f1.append(3)


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1":
                {
                    "a": [1],
                    "ab": [1, 2],
                    "abc": [1, 2, 3]
                }
            },
            {
                "f1":
                {
                    "a": [10],
                    "ab": [2, 1],
                    "abc": [1, 3],
                    "x1": [1, 2, 3]
                }
            }
        )
    ]
)
def test_Dict_List(inp, cmp):
    class R(cnt.Record):
        f1 = cnt.PlainRecordField[cnt.Dict[str, cnt.PlainList[int], tp.Iterable[int]]]()

    r = R(inp)
    assert len(r.f1) == 3

    assert set(r.f1) == {"a", "ab", "abc"}
    assert list(r.f1["a"]) == [1]
    assert list(r.f1["ab"]) == [1, 2]
    assert list(r.f1["abc"][::2]) == [1, 3]

    r.f1["a"][0] = 10
    r.f1["ab"] = [2, 1]
    r.f1["x1"] = r.f1["abc"]
    del r.f1["abc"][1]


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            {
                "f1":
                [
                    {"a": 2}
                ]
            },
            {
                "f1":
                [
                ],
                "f2":
                [
                    {"a": 20, "b": 30},
                    {"c": 40},
                    {"a": 2}
                ]
            }
        )
    ]
)
def test_List_Dict(inp, cmp):
    D1 = cnt.PlainDict[str, int]
    D1x = tp.Mapping[str, int]
    T1 = cnt.RecordField[cnt.List[D1, D1x], tp.Iterable[tp.Union[D1, D1x]]]

    class R(cnt.Record):
        f1 = T1()
        f2 = T1()

    r = R(inp)

    assert len(r.f1) == 1
    assert r.f1[0] == {"a": 2}

    with pytest.raises(AttributeError):
        len(r.f2)

    r.f2.extend(r.f1)
    r.f2.append(r.f1[0])
    r.f1.clear()
    r.f2.insert(1, {"c": 4})
    r.f2[1]["c"] = 40
    with r.f2[0] as x:
        x["b"] = 30
        x["a"] *= 10

    with pytest.raises(AttributeError):
        len(x)


# def test_DirectMutableDictAccessor_subclass():
#     class D(cnt.DirectMutableDictAccessor[str, "cnt.DirectMutableDictAccessor[int, int]"]):
#         _auto_create = True

#     d = {}
#     x = D(d)
#     x["a"][1] = 2

#     assert d == {"a": {1: 2}}


# @pytest.mark.parametrize(
#     'inp,out,tp,value',
#     [
#         ([], [1], int, 1),
#         ([], [""], str, ""),
#         ([], [[]], list, []),
#     ]
# )
# def test_direct_append(inp, cmp, tp, value):
#     x = cnt.DirectMutableListAccessor[tp](inp)
#     x.append(value)
#     cmp(inp)


# @pytest.mark.parametrize(
#     'inp,tp,value',
#     [
#         ([], str, 1),
#         ([], int, ""),
#     ]
# )
# def test_direct_append_wrong_type(inp, tp, value):
#     with pytest.raises(TypeError):
#         x = cnt.DirectMutableListAccessor[tp](inp)
#         x.append(value)


# @pytest.mark.parametrize(
#     'inp,tkey,key,tvalue,value',
#     [
#         ({}, str, 1, str, ""),
#         ({}, int, "", str, ""),
#         ({}, int, 1, str, 1),
#         ({}, int, 1, int, ""),
#         ({}, "int", 1, int, ""),
#     ]
# )
# def test_direct_set_wrong_type(inp, tkey, key, tvalue, value):
#     with pytest.raises(TypeError):
#         x = cnt.DirectMutableListAccessor[tkey, tvalue](inp)
#         x[key] = value


# def test_mutable_namespace():
#     class O1(cnt.DirectMutableObjectAccessor):
#         a: cnt.DictItemMutableListAccessor[int]

#     class O2(cnt.DirectMutableObjectAccessor):
#         b: cnt.DictItemMutableDictAccessor[int, str]

#     class Ns(cnt.MutableNamespace):
#         _paths = {
#             'v': ('x', 'y')
#         }
#         u: O1
#         v: O2

#     d = {}
#     ns = Ns(d)
#     ns.u.a.append(1)
#     ns.v.b[1] = "Z"
#     assert d == {"a": [1], "x": {"y": {"b": {1: "Z"}}}}



import os
from typing import TYPE_CHECKING

from . import _getter
from . import _goplat

if TYPE_CHECKING:
    import pathlib


def make_url(product: str, ver: str) -> str:
    baseurl = os.environ.get("HASHI_RELEASES", "https://releases.hashicorp.com")
    if "-dev" in ver:
        baseurl = os.environ.get("HASHI_DEV_RELEASES", baseurl)

    return "{baseurl}/{product}/{ver}/{product}_{ver}_{goos}_{goarch}.zip".format(
        baseurl=baseurl,
        product=product,
        ver=ver,
        goos=_goplat.goos(),
        goarch=_goplat.goarch(),
    )


def get_hashi_product(product: str, ver: str) -> "pathlib.Path":
    return _getter.get(make_url(product, ver))


def get_terraform(ver: str) -> "pathlib.Path":
    return get_hashi_product("terraform", ver)


def get_terraform_provider(name: str, ver: str) -> "pathlib.Path":
    return get_hashi_product("terraform-provider-{}".format(name), ver)



import sys
import textwrap
from typing import Set

import jedi
import pytest

import yapytf._containers2 as cnt


prefix_code = textwrap.dedent("""
    import yapytf
    import yapytf._containers2 as cnt

    """)


# new in Python 3.7, not yet added to typeshed
STR_IGNORE = {"isascii"}


@pytest.mark.parametrize(
    'code,ref,ignore',
    [
        (
            """
            x = dict()
            x.
            """,
            dir(dict),
            set()
        ),
        (
            """
            x = ""
            x.
            """,
            dir(str),
            STR_IGNORE
        ),
        (
            """
            from typing import Mapping
            x: Mapping[int, str]
            x[0].
            """,
            dir(str),
            STR_IGNORE
        ),
        (
            """
            cnt.
            """,
            dir(cnt),
            set()
        ),
        (
            """
            class R(cnt.Record):
                f1 = cnt.ConstRecordField[str]()

            r = R()
            r.
            """,
            ["f1"],
            set()
        ),
        (
            """
            class R(cnt.Record):
                f1 = cnt.ConstRecordField[str]()

            r = R()
            r.f1.
            """,
            dir(str),
            STR_IGNORE
        ),
        (
            """
            class R(cnt.Record):
                f1 = cnt.ConstRecordField[cnt.PlainDict[int, str]]()

            r = R()
            r.f1[0].
            """,
            dir(str),
            STR_IGNORE
        ),
        (
            """
            class R(cnt.Record):
                f1 = cnt.ConstRecordField[cnt.PlainDict[int, str]]()

            r = R()
            with r as x:
                x.
            """,
            ["f1"],
            set()
        ),
        (
            """
            class R(cnt.Record):
                f1 = cnt.ConstRecordField[cnt.PlainDict[int, str]]()

            r = R()
            r.f1.
            """,
            dir(dict),
            {"copy", "fromkeys"}
        ),
    ]
)
def test(code: str, ref: str, ignore: Set[str]) -> None:
    code = prefix_code + textwrap.dedent(code)
    lines = code.splitlines()
    script = jedi.Script(
        source=code,
        line=len(lines),
        column=len(lines[-1]),
        path="source.py",
        # sys_path=sys.path
    )

    def nounders(l):
        return (i for i in l if not i.startswith("_"))

    completions = set(nounders(i.name for i in script.completions())) - ignore
    expected_completions = set(nounders(ref)) - ignore
    assert completions == expected_completions



import pathlib as _pathlib
import typing as _typing

from implements import Interface as _Interface

JsonType = _typing.Dict[str, _typing.Any]
SimpleOutputResult = _typing.Dict[str, _typing.Union[str, _typing.Tuple[str, int]]]


class StateBackendConfig:
    __slots__ = ('name', 'vars')

    def __init__(self) -> None:
        self.name: str = "local"
        self.vars: _typing.Dict[str, str] = {}


class Configurator(_Interface):
    @staticmethod
    def requires() -> _typing.Set[str]:
        return set()

    @classmethod
    def schema(cls, schema: JsonType) -> None:
        for i in cls.schema_required_strings():
            schema["properties"].update({i: {"type": "string"}})
            schema["required"].append(i)

    @staticmethod
    def schema_required_strings() -> _typing.Set[str]:
        return set()

    @staticmethod
    def versions(versions: _typing.Dict[str, _typing.Any]) -> None:
        pass

    def __init__(self, data: JsonType) -> None:
        self._data = data

    @property
    def data(self) -> JsonType:
        return self._data

    def state_backend_cfg(self, cfg: StateBackendConfig) -> None:
        pass

    def populate_providers(
        self,
        *,
        model: "yapytfgen.providers_model",  # type: ignore  # noqa
    ) -> None:
        pass

    def populate(
        self,
        *,
        model: "yapytfgen.model",  # type: ignore  # noqa
        step_data: _typing.Any,
    ) -> None:
        pass

    def output(
        self,
        *,
        state: "yapytfgen.state",  # type: ignore  # noqa
        dest: _pathlib.Path,
    ) -> None:
        for path_str, value in self.simple_output(state=state).items():
            mode: _typing.Optional[int]

            if isinstance(value, tuple):
                text, mode = value
            else:
                text, mode = value, None

            path = _pathlib.PurePosixPath(path_str)

            if path.is_absolute():
                raise RuntimeError(f"simple_output returned absolute path: {path_str}")

            if ".." in path.parts:
                raise RuntimeError(f"simple_output returned path containing \"..\": {path_str}")

            file_path = dest.joinpath(path)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(text)

            if mode is not None:
                file_path.chmod(mode)

    def simple_output(
        self,
        *,
        state: "yapytfgen.state",  # type: ignore  # noqa
    ) -> SimpleOutputResult:
        return {}



import abc as _abc
import copy as _copy
import dataclasses as _dataclasses
import sys as _sys
import typing as _tp
import typeguard as _typeguard
import typing_inspect as _typing_inspect


_CONCRETE_CONTAINERS: _tp.Dict[_tp.Type[_tp.Any], _abc.ABCMeta] = {
    _tp.Dict: dict,
    _tp.List: list,
}


def _get_generic_types(
    instance: _tp.Any,
    count: _tp.Union[int, _tp.Set[int]],
    *,
    hint: str = ""
) -> _tp.Tuple[_tp.Type[_tp.Any], ...]:
    types = _typing_inspect.get_args(_typing_inspect.get_generic_type(instance))

    if not types:
        types = _typing_inspect.get_args(_typing_inspect.get_generic_bases(instance)[0])

    globalns = _sys.modules[instance.__class__.__module__].__dict__

    _eval_type = _tp._eval_type  # type: ignore
    types = tuple(_eval_type(i, globalns, None) for i in types)

    if isinstance(count, int):
        count = {count}

    if count != {-1} and len(types) not in count or any(_typing_inspect.is_typevar(i) for i in types):
        raise TypeError(f"{instance.__class__.__name__} generic was not properly parameterized{hint}: {types}")

    return types


class _Never:
    pass


_Key = _tp.Optional[_tp.Union[str, int]]
_DictAny = _tp.Dict[_tp.Any, _tp.Any]
_ListAny = _tp.List[_tp.Any]


@_dataclasses.dataclass
class _PathComponent:
    name: _tp.Any
    is_attr: bool


def _traverse_dict(d: _DictAny, path: _tp.Iterable[_PathComponent]) -> _tp.Any:
    result = d

    for component in path:
        try:
            result = result[component.name]
        except KeyError as e:
            if component.is_attr:
                raise AttributeError(e)
            else:
                raise

    return result


T = _tp.TypeVar("T")


class _Getter(_abc.ABC, _tp.Generic[T]):
    @_abc.abstractmethod
    def get(self) -> T:
        raise NotImplementedError

    @_abc.abstractmethod
    def get_autocreate(self) -> T:
        raise NotImplementedError


class _DirectGetter(_Getter[T]):
    __slots__ = ("_val")

    def __init__(self, val: T) -> None:
        self._val = val

    def get(self) -> T:
        return self._val

    def get_autocreate(self) -> T:
        return self._val


class _NestedDictGetter(_Getter[T]):
    __slots__ = ("_dict", "_path", "_type", "_default")

    def __init__(
        self,
        src: _tp.Union[_DictAny, "_Getter[_tp.Any]"],
        path: _tp.Iterable[_PathComponent],
        default: _tp.Type[T]
    ) -> None:
        path = list(path)

        if isinstance(src, _NestedDictGetter):
            path = src._path + path
            src = src._dict
        elif isinstance(src, _Getter):
            src = src.get_autocreate()

        assert isinstance(src, dict)

        self._dict: _DictAny = src
        self._path: _tp.List[_PathComponent] = path
        self._type = default
        concrete = _tp.cast(_tp.Type[T], _typing_inspect.get_origin(default)) or default
        self._default = _CONCRETE_CONTAINERS.get(concrete, concrete)

    def get(self) -> T:
        result = _traverse_dict(self._dict, self._path)
        _typeguard.check_type("", result, self._type)
        return _tp.cast(T, result)

    def get_autocreate(self) -> T:
        result = self._dict

        if self._path:
            for component in self._path[:-1]:
                result = result.setdefault(component.name, {})

            result = result.setdefault(self._path[-1].name, self._default())

        _typeguard.check_type("", result, self._default)
        return _tp.cast(T, result)


Tself = _tp.TypeVar("Tself", bound="_ContextManagerMixin")


class _ContextManagerMixin:
    def __init__(self) -> None:
        self._context_copy: _tp.Optional["_ContextManagerMixin"] = None

    def _reset_getter(self) -> None:
        raise NotImplementedError

    def __enter__(self: Tself) -> Tself:
        assert self._context_copy is None
        self._context_copy = _copy.copy(self)
        return self._context_copy

    def __exit__(
        self,
        *args: _tp.Any,
    ) -> None:
        assert self._context_copy is not None
        self._context_copy._reset_getter()
        self._context_copy = None


class Record(_ContextManagerMixin):
    __slots__ = ("_getter")

    _getter: _Getter[_DictAny]

    def __init__(
        self,
        container: _tp.Union[_DictAny, _Getter[_DictAny]],
        # *,
        # path: _tp.Iterable[_tp.Any] = []
    ) -> None:
        _ContextManagerMixin.__init__(self)
        _typeguard.check_argument_types()
        path: _tp.Iterable[_tp.Any] = []

        if isinstance(container, _Getter):
            assert not path
            self._getter = container
        else:
            _typeguard.check_type("container", container, dict)
            self._getter = _NestedDictGetter(container, (_PathComponent(i, False) for i in path), _DictAny)

    def _reset_getter(self) -> None:
        self._getter = None  # type: ignore


Tsetter = _tp.TypeVar("Tsetter")
Tval = _tp.TypeVar("Tval")
Tkey = _tp.TypeVar("Tkey", bound=_Key)


def _generic_to_tuple(
    type_: _tp.Type[_tp.Any]
) -> _tp.Union[_tp.Type[_tp.Any], _tp.Tuple[_tp.Type[_tp.Any], ...]]:
    if _typing_inspect.is_generic_type(type_):
        class_ = _typing_inspect.get_origin(type_)
        types = _typing_inspect.get_args(type_)
        return (class_, *types)

    return type_


_TConstCtor = _tp.Callable[[_tp.Any], _tp.Any]


def _make_const_ctor(type_: _tp.Type[_tp.Any]) -> _TConstCtor:
    type2 = _generic_to_tuple(type_)

    if isinstance(type2, tuple):
        class_ = type2[0]
        types = type2[1:]

        if class_ is None:

            if getattr(type_, "_is_protocol", False):

                def ctor(raw: _tp.Any) -> _tp.Any:
                    _typeguard.check_type("value", raw, type_)
                    return raw

            else:

                def ctor(raw: _tp.Any) -> _tp.Any:
                    return type_(raw)

        else:

            def ctor(raw: _tp.Any) -> _tp.Any:
                return class_(raw, types)

    elif issubclass(type_, Record):

        def ctor(raw: _tp.Any) -> _tp.Any:
            return type_(raw)

    else:

        def ctor(raw: _tp.Any) -> _tp.Any:
            _typeguard.check_type("value", raw, type_)
            return raw

    return ctor


class _ConstRecordField(_tp.Generic[Tval]):
    __slots__ = ("path", "name", "val_ctor", "no_name")

    def __init__(
        self,
        *,
        path: _tp.List[str] = [],
        name: _tp.Optional[str] = None,
        no_name: bool = False,
    ) -> None:
        self.path = path
        self.name = name
        self.no_name = no_name
        self.val_ctor: _tp.Optional[_TConstCtor] = None

    def __set_name__(self, owner: _tp.Type[Record], name: str) -> None:
        if self.name is None:
            self.name = name

    def __set__(self, instance: Record, value: _Never) -> None:
        raise NotImplementedError

    def __delete__(self, instance: Record) -> None:
        raise NotImplementedError

    def __get__(self, instance: Record, owner: _tp.Type[Record]) -> Tval:
        assert self.name is not None

        container = instance._getter.get()

        for i in self.path:
            container = container.setdefault(i, {})

        if self.no_name:
            v = container
        else:
            try:
                v = container[self.name]
            except KeyError as e:
                raise AttributeError(e)

        if self.val_ctor is None:
            self.val_ctor = _make_const_ctor(_get_generic_types(self, 1, hint=f". Field {self.name}")[0])

        return _tp.cast(Tval, self.val_ctor(v))


class ConstRecordField(_ConstRecordField[T]):
    pass


class ConstDict(_tp.Mapping[Tkey, Tval], _ContextManagerMixin):
    __slots__ = ("_container", "_key_type", "_val_ctor")

    def __init__(self, container: _DictAny, types: _tp.Tuple[_tp.Type[_tp.Any], ...]) -> None:
        _ContextManagerMixin.__init__(self)
        _typeguard.check_type("container", container, dict)
        self._container = container
        self._key_type, val_type = types
        self._val_ctor = _make_const_ctor(val_type)

    def _reset_getter(self) -> None:
        self._container = None  # type: ignore

    def __len__(self) -> int:
        return len(self._container)

    def __getitem__(self, k: Tkey) -> Tval:
        _typeguard.check_type("key", k, self._key_type)
        return _tp.cast(Tval, self._val_ctor(self._container[k]))

    def __iter__(self) -> _tp.Iterator[Tkey]:
        def check_key(k: _tp.Any) -> Tkey:
            _typeguard.check_type("key", k, self._key_type)
            return _tp.cast(Tkey, k)

        return (check_key(i) for i in self._container.__iter__())


class ConstList(_tp.Sequence[T], _ContextManagerMixin):
    __slots__ = ("_container", "_type_ctor")

    def __init__(self, container: _tp.List[_tp.Any], types: _tp.Tuple[_tp.Type[_tp.Any], ...]) -> None:
        _ContextManagerMixin.__init__(self)
        self._container = container
        type_, = types
        self._type_ctor = _make_const_ctor(type_)

    def _reset_getter(self) -> None:
        self._container = None  # type: ignore

    def __len__(self) -> int:
        return len(self._container)

    @_tp.overload
    def __getitem__(self, i: int) -> T:
        ...

    @_tp.overload
    def __getitem__(self, i: slice) -> _tp.Sequence[T]:
        ...

    def __getitem__(self, i: _tp.Union[int, slice]) -> _tp.Union[T, _tp.Sequence[T]]:
        def validate(idx: int, v: _tp.Any) -> T:
            return _tp.cast(T, self._type_ctor(v))

        if isinstance(i, slice):
            start, stop, stride = i.indices(len(self._container))
            return [
                validate(start + stride * j, v)
                for j, v in enumerate(self._container[i])
            ]
        else:
            return validate(i, self._container[i])


def _make_ctor(
    type_: _tp.Type[_tp.Any],
    is_attr: bool
) -> _tp.Callable[[_Getter[_DictAny], _tp.Any], _tp.Any]:
    type2 = _generic_to_tuple(type_)

    if isinstance(type2, tuple):
        class_ = type2[0]
        types = type2[1:]

        def ctor(getter: _Getter[_DictAny], key: _tp.Any) -> _tp.Any:
            return class_(
                _NestedDictGetter(
                    getter,
                    [_PathComponent(key, is_attr)],
                    class_._container_type()
                ),
                types
            )

    elif issubclass(type_, Record):

        rec_class: _tp.Type[Record] = type_

        def ctor(getter: _Getter[_DictAny], key: _tp.Any) -> _tp.Any:
            return rec_class(
                _NestedDictGetter(
                    getter,
                    [_PathComponent(key, is_attr)],
                    _DictAny
                )
            )

    else:

        def ctor(getter: _Getter[_DictAny], key: _tp.Any) -> _tp.Any:
            result = getter.get()[key]
            _typeguard.check_type(key, result, type_)
            return result

    return ctor


def _copy_into(getter: _Getter[_DictAny], key: _tp.Any, val: _tp.Any) -> None:
    getter.get_autocreate()[key] = _copy.deepcopy(val._getter.get())


def _make_setter(
    type_: _tp.Type[_tp.Any],
    is_attr: bool
) -> _tp.Callable[[_Getter[_DictAny], _tp.Any, _tp.Any], None]:
    type2 = _generic_to_tuple(type_)

    if isinstance(type2, tuple):
        class_ = type2[0]
        types = type2[1:]

        def setter(getter: _Getter[_DictAny], key: _tp.Any, val: _tp.Any) -> None:
            if isinstance(val, class_):
                _copy_into(getter, key, val)
            else:
                obj = class_(
                    _NestedDictGetter(
                        getter,
                        [_PathComponent(key, is_attr)],
                        class_._container_type()
                    ),
                    types
                )
                obj._populate(val)

    elif issubclass(type_, Record):

        rec_class: _tp.Type[Record] = type_

        def setter(getter: _Getter[_DictAny], key: _tp.Any, val: _tp.Any) -> None:
            _typeguard.check_type("value", val, rec_class)
            _copy_into(getter, key, val)

    else:

        def setter(getter: _Getter[_DictAny], key: _tp.Any, val: _tp.Any) -> None:
            _typeguard.check_type("value", val, type_)
            getter.get_autocreate()[key] = val

    return setter


class _RecordField(_tp.Generic[Tval, Tsetter]):
    __slots__ = (
        "name",
        "val_ctor",
        "val_setter",
        "val_type",
        "val_setter_type"
    )

    def __init__(self, *, name: _tp.Optional[str] = None) -> None:
        self.name = name
        self.val_type: _tp.Optional[_tp.Type[_tp.Any]] = None

    def __set_name__(self, owner: _tp.Type[Record], name: str) -> None:
        if self.name is None:
            self.name = name

    def _lazy_init(self) -> None:
        if self.val_type is None:
            types = _get_generic_types(self, {1, 2}, hint=f". Field {self.name}")
            if len(types) == 1:
                types = types * 2
            self.val_type, self.val_setter_type = types
            self.val_ctor = _make_ctor(self.val_type, True)
            self.val_setter = _make_setter(self.val_type, True)

    def __set__(self, instance: Record, value: Tsetter) -> None:
        assert self.name is not None
        self._lazy_init()
        _typeguard.check_type("value", value, self.val_setter_type)
        self.val_setter(instance._getter, self.name, value)

    def __delete__(self, instance: Record) -> None:
        assert self.name is not None
        container = instance._getter.get()
        try:
            del container[self.name]
        except KeyError as e:
            raise AttributeError(e)

    def __get__(self, instance: Record, owner: _tp.Type[Record]) -> Tval:
        assert self.name is not None
        self._lazy_init()
        try:
            return _tp.cast(Tval, self.val_ctor(instance._getter, self.name))
        except KeyError as e:
            raise AttributeError(e)


class RecordField(_RecordField[Tval, Tsetter]):
    pass


class PlainRecordField(_RecordField[Tval, Tval]):
    pass


class _Dict(_tp.MutableMapping[Tkey, Tval], _tp.Generic[Tkey, Tval, Tsetter], _ContextManagerMixin):
    __slots__ = (
        "_getter",
        "_key_type",
        "_val_ctor",
        "_val_setter",
        "_val_setter_type"
    )

    @staticmethod
    def _container_type() -> _tp.Type[_tp.Any]:
        return dict

    def __init__(
        self,
        getter: _Getter[_DictAny],
        types: _tp.Tuple[_tp.Type[_tp.Any], ...],
    ) -> None:
        _ContextManagerMixin.__init__(self)
        self._getter = getter

        if len(types) == 2:
            types = types + (types[1],)

        self._key_type, val_type, self._val_setter_type = types
        self._val_ctor = _make_ctor(val_type, False)
        self._val_setter = _make_setter(val_type, False)

    def _reset_getter(self) -> None:
        self._getter = None  # type: ignore

    def _validate_key(self, k: Tkey) -> None:
        _typeguard.check_type("key", k, self._key_type)

    def __len__(self) -> int:
        return len(self._getter.get())

    def __getitem__(self, k: Tkey) -> Tval:
        self._validate_key(k)
        return _tp.cast(Tval, self._val_ctor(self._getter, k))

    def __iter__(self) -> _tp.Iterator[Tkey]:
        def check_key(k: _tp.Any) -> Tkey:
            self._validate_key(k)
            return _tp.cast(Tkey, k)

        return (check_key(i) for i in self._getter.get().__iter__())

    # ignore Argument 2 of "__setitem__" incompatible with supertype "MutableMapping"
    # which is due to Tval/Tsetter duality

    def __setitem__(self, k: Tkey, v: Tsetter) -> None:  # type: ignore
        self._validate_key(k)
        _typeguard.check_type("value", v, self._val_setter_type)
        self._val_setter(self._getter, k, v)

    def __delitem__(self, k: Tkey) -> None:
        self._validate_key(k)
        self._getter.get().__delitem__(k)

    def _populate(self, val: _tp.Any) -> None:
        _typeguard.check_type("value", val, dict)
        self.clear()
        for k, v in val.items():
            self[k] = v


class Dict(_Dict[Tkey, Tval, Tsetter]):
    pass


class PlainDict(_Dict[Tkey, Tval, Tval]):
    pass


class _List(_tp.MutableSequence[Tval], _tp.Generic[Tval, Tsetter], _ContextManagerMixin):
    __slots__ = (
        "_getter",
        "_val_ctor",
        "_val_setter",
        "_val_type",
        "_val_setter_type"
    )

    @staticmethod
    def _container_type() -> _tp.Type[_tp.Any]:
        return list

    def __init__(
        self,
        getter: _Getter[_ListAny],
        types: _tp.Tuple[_tp.Type[_tp.Any], ...],
    ) -> None:
        _ContextManagerMixin.__init__(self)
        self._getter = getter

        if len(types) == 1:
            types = types * 2

        self._val_type, self._val_setter_type = types

        type_ = _generic_to_tuple(self._val_type)

        def cast(val: _tp.Any) -> Tval:
            return _tp.cast(Tval, val)

        if isinstance(type_, tuple):
            class_ = type_[0]
            types = type_[1:]

            def check_type(i: int, raw: _tp.Any) -> None:
                _typeguard.check_type(f"[{i}]", raw, class_._container_type())

            if issubclass(class_, _List):

                def ctor(i: int, raw: _tp.Any) -> Tval:
                    check_type(i, raw)
                    return cast(class_(_DirectGetter(raw), types))

            else:

                def ctor(i: int, raw: _tp.Any) -> Tval:
                    check_type(i, raw)
                    return cast(class_(_NestedDictGetter(raw, [], class_._container_type()), types))

            def setter(val: _tp.Any) -> _tp.Any:
                if isinstance(val, class_):
                    return _copy.deepcopy(val._getter.get())
                else:
                    result = class_._container_type()()
                    obj = class_(_DirectGetter(result), types)
                    obj._populate(val)
                    return result

        elif issubclass(type_, Record):

            rec_class: _tp.Type[Record] = type_

            def ctor(i: int, raw: _tp.Any) -> Tval:
                _typeguard.check_type(f"[{i}]", raw, dict)
                return cast(rec_class(_NestedDictGetter(raw, [], _DictAny)))

            def setter(val: _tp.Any) -> _tp.Any:
                _typeguard.check_type("value", val, type_)
                return _copy.deepcopy(val._getter.get())

        else:

            def ctor(i: int, raw: _tp.Any) -> Tval:
                _typeguard.check_type(f"[{i}]", raw, type_)
                return cast(raw)

            def setter(val: _tp.Any) -> _tp.Any:
                _typeguard.check_type("value", val, type_)
                return _copy.deepcopy(val)

        self._val_ctor = ctor
        self._val_setter = setter

    def _reset_getter(self) -> None:
        self._getter = None  # type: ignore

    def _populate(self, val: _tp.Any) -> None:
        _typeguard.check_type("value", val, list)
        self._getter.get_autocreate().clear()
        for i in val:
            self.append(i)

    def __len__(self) -> int:
        return len(self._getter.get())

    @_tp.overload
    def __getitem__(self, i: int) -> Tval:
        ...

    @_tp.overload
    def __getitem__(self, i: slice) -> _tp.MutableSequence[Tval]:
        ...

    def __getitem__(self, i: _tp.Union[int, slice]) -> _tp.Union[Tval, _tp.MutableSequence[Tval]]:
        container = self._getter.get()

        if isinstance(i, slice):
            start, stop, stride = i.indices(len(container))
            return [
                self._val_ctor(start + stride * j, v)
                for j, v in enumerate(container[i])
            ]
        else:
            return self._val_ctor(i, container[i])

    # ignore Signature of "__setitem__" incompatible with supertype "MutableSequence"
    # which is due to Tval/Tsetter duality

    @_tp.overload  # type: ignore
    def __setitem__(self, i: int, v: Tsetter) -> None:
        ...

    @_tp.overload
    def __setitem__(self, i: slice, v: _tp.Iterable[Tsetter]) -> None:
        ...

    def __setitem__(self, i: _tp.Union[int, slice], v: _tp.Union[T, _tp.Iterable[Tsetter]]) -> None:
        container = self._getter.get()
        if isinstance(i, slice):
            container[i] = (self._val_setter(j) for j in _tp.cast(_tp.Iterable[Tsetter], v))
        else:
            container[i] = self._val_setter(v)

    @_tp.overload
    def __delitem__(self, i: int) -> None:
        ...

    @_tp.overload
    def __delitem__(self, i: slice) -> None:
        ...

    def __delitem__(self, i: _tp.Union[int, slice]) -> None:
        self._getter.get().__delitem__(i)

    # ignore Argument 2 of "insert" incompatible with supertype "MutableSequence"
    # which is due to Tval/Tsetter duality

    def insert(self, i: int, v: Tsetter) -> None:  # type: ignore
        self._getter.get().insert(i, self._val_setter(v))

    # ignore Argument 1 of "append" incompatible with supertype "MutableSequence"
    # which is due to Tval/Tsetter duality

    def append(self, v: Tsetter) -> None:  # type: ignore
        # note: a separate "append" is needed as default one does "len" before
        # self._getter.get_autocreate() is called, causing dict item autocreation to fail

        self._getter.get_autocreate().append(self._val_setter(v))

    def insert_new(self, i: int) -> T:
        pass
        # TODO
        # self._get_container_for_write().insert(i, self._validate_and_copy_value(0, v))


class List(_List[Tval, Tsetter]):
    pass


class PlainList(_List[Tval, Tval]):
    pass



from typing import Any, Dict, Tuple


def get_resources_attrs(
    j: Dict[str, Any],
    resources_schemas_versions: Dict[Tuple[str, str], int]
) -> Dict[str, Any]:
    ver = j["version"]
    if not ver == 4:
        raise RuntimeError(f"Unsupported version of terraform state: \"{ver}\"")

    result: Dict[str, Any] = {}

    for res in j["resources"]:
        res_mode = res["mode"]
        res_type = res["type"]
        res_name = res["name"]
        k = result.setdefault(res_mode, {}).setdefault(res_type, {}).setdefault(res_name, {})
        assert not k
        for instance in res["instances"]:
            index_key = instance.get("index_key")

            instance_schema_version = instance["schema_version"]
            expected_instance_schema_version = resources_schemas_versions[res_mode, res_type]
            if instance_schema_version != expected_instance_schema_version:
                raise RuntimeError(
                    "Unexpected resource schema version for {}.{}.{}{}: got {}, expected {}".format(
                        res_mode,
                        res_type,
                        res_name,
                        f"[{index_key}]" if index_key else "",
                        instance_schema_version,
                        expected_instance_schema_version
                    )
                )

            assert index_key not in k
            k[index_key] = instance["attributes"]

    return result



import ast
import keyword
import pathlib
from typing import (Any, Callable, Dict, Generator, Iterable, List, Mapping,
                    Optional, Tuple, Union)

from . import _pcache

# DATA_TYPE_HINT = "_tp.Dict[str, _tp.Any]"
KIND_TO_KEY = {
    "data_source": "data",
    "resource": "resource",
}
STATE_KIND_TO_KEY = {
    "data_source": "data",
    "resource": "managed",
}

# https://www.terraform.io/docs/configuration/resources.html#meta-arguments
# TODO: this is incomplete

_RES_META_ARGS_SCHEMA = {
    "block": {
        "attributes": {
            "depends_on": {
                "type": ["set", "string"],
                "optional": True
            },
            "count": {
                "type": "number",
                "optional": True
            },
            "provider": {
                "type": "string",
                "optional": True
            },
        }
    }
}

_COMMON_IMPORTS = [
    "import typing as _tp",
    "from yapytf import _containers as _c",
    "from yapytf import _genbase",
]

_BuilderChunkGeneratorType = Generator[Tuple[int, str], None, None]


class Builder:
    def __init__(self) -> None:
        self.chunks: List[Callable[[str], _BuilderChunkGeneratorType]] = []
        self.indent_level = 0

    def produce(self, indent: str = "    ") -> str:
        result = []
        blanks = 0

        for chunk_producer in self.chunks:
            for n, s in chunk_producer(indent):
                blanks = max(blanks, n)

                if s:
                    for _ in range(blanks):
                        result.append("\n")

                    blanks = 0
                    result.append(s)

        return "".join(result)

    def blanks(self, n: int) -> None:
        def produce(indent: str) -> _BuilderChunkGeneratorType:
            yield n, ""

        if n:
            self.chunks.append(produce)

    def line(self, s: str) -> None:
        def produce(indent: str) -> _BuilderChunkGeneratorType:
            if s:
                tabs = 0

                while s[tabs:tabs + 1] == "\t":
                    tabs += 1

                for i in range(self.indent_level + tabs):
                    yield 0, indent

                yield 0, s[tabs:]

            yield 0, "\n"

        self.chunks.append(produce)

    def lines(self, lines: List[str]) -> None:
        for line in lines:
            self.line(line)

    def block(self, indented: bool = True) -> "Builder":
        block_builder = Builder()
        block_builder.indent_level = self.indent_level + int(indented)

        def produce(indent: str) -> _BuilderChunkGeneratorType:
            for chunk_producer in block_builder.chunks:
                yield from chunk_producer(indent)

        self.chunks.append(produce)

        return block_builder


def def_block(
    builder: Builder,
    blanks: Union[int, Tuple[int, int]],
    prefix: str,
    items: Optional[List[str]] = None,
    suffix: str = "",
    *,
    decorators: List[str] = [],
    lines: List[str] = [],
    doc_string: Optional[str] = None,
) -> Builder:
    if isinstance(blanks, int):
        blanks_before, blanks_after = blanks, blanks
    else:
        blanks_before, blanks_after = blanks

    if suffix:
        suffix = f" -> {suffix}"

    builder.blanks(blanks_before)

    for i in decorators:
        builder.line(f"@{i}")

    if items is None:
        builder.line(f"{prefix}{suffix}:")
    elif len(items) == 1:
        builder.line(f"{prefix}({items[0]}){suffix}:")
    else:
        builder.line(f"{prefix}(")
        items_block = builder.block()
        for item in items:
            items_block.line(f"{item},")
        builder.line(f"){suffix}:")

    block1 = builder.block()

    if blanks_after:
        block2 = block1.block(indented=False)
        block1.blanks(blanks_after)
        result = block2
    else:
        result = block1

    if doc_string is not None:
        result.lines(['"""', doc_string, '"""'])

    result.lines(lines)

    return result


def join_class_path(path: List[str], name: str) -> str:
    return "\"{}\"".format(".".join(path + [name]))


def make_ns_class(
    *,
    builder: Builder,
    class_name: str,
    props: Mapping[str, str],
    data_paths: Mapping[str, List[str]] = {},
    nested: bool = False,
) -> None:
    class_builder = def_block(
        builder,
        1 if nested else 2,
        f"class {class_name}",
        ["_c.Record"],
        doc_string="",
    )

    for prop_name, prop_type in sorted(props.items()):
        assert prop_name.isidentifier()
        prop_name_slug = "_" if keyword.iskeyword(prop_name) else ""

        extra = ""

        data_path = data_paths.get(prop_name, [])
        if data_path:
            extra += f", path={repr(data_path)}"

        class_builder.line(f"{prop_name}{prop_name_slug} = _c.ConstRecordField[{prop_type}](no_name=True{extra})")


def build_record_field(
    tf_type: Any,
    *,
    reader: bool,
    builder: Builder,
    attr_name: str,
    class_path: List[str],
) -> str:
    def inner(
        tf_type: Any,
        *,
        depth: int = 0,
    ) -> Union[str, Tuple[str, str]]:
        if isinstance(tf_type, list):
            tf_type_kind, tf_type_inner = tf_type

            if tf_type_kind in {"list", "set", "map"}:
                child = inner(tf_type_inner, depth=depth + 1)

                if reader:
                    if tf_type_kind == "map":
                        return f"_c.ConstDict[str, {child}]"
                    else:
                        return f"_c.ConstList[{child}]"
                else:
                    ntg = f"_{attr_name}_g{depth}"
                    nts = f"_{attr_name}_s{depth}"

                    if isinstance(child, str):
                        t1 = "Plain"
                        t3 = f"{child}"
                        t4 = child
                    else:
                        t1 = ""
                        t3 = f"{child[0]}, {child[1]}"
                        t4 = child[1]

                    if tf_type_kind == "map":
                        t2 = "Dict[str, "
                        ts = f"_tp.Mapping[str, {t4}]"
                    else:
                        t2 = "List["
                        ts = f"_tp.Sequence[{t4}]"

                    builder.line(f"{ntg} = _c.{t1}{t2}{t3}]")
                    builder.line(f"{nts} = _tp.Union[{ntg}, {ts}]")

                    return ntg, nts

            if tf_type_kind == "object":
                assert isinstance(tf_type_inner, dict)

                instance_class_name = f"_{attr_name}_type"
                full_instance_class_name = join_class_path(class_path, instance_class_name)

                class_builder = def_block(
                    builder,
                    1,
                    f"class {instance_class_name}",
                    ["_c.Record"]
                )

                for item_name, item_tf_type in sorted(tf_type_inner.items()):
                    assert item_name.isidentifier()

                    field_params: List[str] = []

                    slug = "_" if keyword.iskeyword(item_name) else ""
                    if slug:
                        field_params.append(f"name=\"{item_name}\"")

                    child = build_record_field(
                        item_tf_type,
                        reader=reader,
                        attr_name=item_name,
                        class_path=class_path + [instance_class_name],
                        builder=class_builder,
                    )

                    class_builder.line(f"{item_name}{slug} = {child}({', '.join(field_params)})")

                if not tf_type_inner.items():
                    class_builder.line("pass")

                return full_instance_class_name
        else:
            if tf_type == "bool":
                return "bool"
            if tf_type == "string":
                return "str"
            if tf_type == "number":
                return "int"

        assert 0, f"Unknown Terraform type {tf_type}"

    r = inner(tf_type)

    if reader:
        assert isinstance(r, str)
        return f"_c.ConstRecordField[{r}]"
    else:
        if isinstance(r, str):
            return f"_c.PlainRecordField[{r}]"
        else:
            return f"_c.RecordField[{r[0]}, {r[1]}]"


def make_schema_class(
    *,
    builder: Builder,
    class_name: str,
    schemas: Iterable[Mapping[str, Any]],
    class_path: List[str],
    reader: bool,
) -> str:
    full_class_name = join_class_path(class_path, class_name)

    class_builder = def_block(
        builder,
        1 if class_path else 2,
        f"class {class_name}",
        ["_c.Record"]
    )

    empty = True

    for schema in schemas:
        for attr_name, attr_schema in sorted(schema.get("block", {}).get("attributes", {}).items()):
            attr_optional = attr_schema.get("optional", False)
            attr_computed = attr_schema.get("computed", False)
            if not(reader or not(attr_computed) or attr_optional):
                continue

            assert attr_name.isidentifier()
            attr_slug = "_" if keyword.iskeyword(attr_name) else ""

            python_type = build_record_field(
                attr_schema["type"],
                builder=class_builder,
                reader=reader,
                attr_name=attr_name,
                class_path=class_path + [class_name],
            )

            field_params = []
            if attr_slug:
                field_params.append(f"name=\"{attr_name}\"")

            class_builder.line(f"{attr_name}{attr_slug} = {python_type}({', '.join(field_params)})")
            empty = False

        for attr_name, attr_schema in sorted(schema.get("block", {}).get("block_types", {}).items()):
            assert attr_name.isidentifier()
            attr_slug = "_" if keyword.iskeyword(attr_name) else ""

            attr_class_name = make_schema_class(
                builder=class_builder,
                class_name=f"_{attr_name}_type",
                schemas=[attr_schema],
                class_path=class_path + [class_name],
                reader=reader,
            )

            attr_nesting_mode = attr_schema["nesting_mode"]
            if attr_nesting_mode == "single":
                class_builder.line(f"{attr_name}{attr_slug}: {attr_class_name}")
            elif attr_nesting_mode in {"list", "set"}:
                # TODO "if reader?"
                class_builder.line(f"{attr_name}{attr_slug}: _c.PlainList[{attr_class_name}]  # xxx")
            else:
                assert 0, f"Unknown Terraform nesting_mode {attr_nesting_mode}"

    if empty:
        class_builder.line("pass")

    return full_class_name


def gen_provider_py(
    *,
    work_dir: pathlib.Path,
    terraform_version: str,
    provider_name: str,
    provider_version: str,
    provider_schema: Dict[str, Any],
) -> pathlib.Path:
    key = "provider-py-{}-{}-{}".format(
        terraform_version, provider_name, provider_version
    )

    def produce(dir_path: pathlib.Path) -> None:
        def finalize_builder(name: str, builder: Builder) -> None:
            fname = f"{name}.py"
            produced = builder.produce()
            try:
                ast.parse(produced, filename=fname)
            except SyntaxError:
                work_dir.joinpath(fname).write_text(produced)
                raise

            dir_path.joinpath(fname).write_text(produced)

        provider_name_ = f"{provider_name}_"

        pschema = provider_schema["provider_schemas"][provider_name]
        builder = Builder()
        imports_block = builder.block(indented=False)
        imports_block.lines(_COMMON_IMPORTS)

        def make_v1() -> None:
            make_schema_class(
                builder=builder,
                class_name="v1_model_provider",
                schemas=[pschema["provider"]],
                class_path=[],
                reader=False,
            )

            for kind in ["data_source", "resource"]:
                ns_props: Dict[str, Dict[str, str]] = {}
                ns_paths: Dict[str, Dict[str, List[str]]] = {}

                for rname, rschema in sorted(pschema.get(f"{kind}_schemas", {}).items()):
                    if rname == provider_name:
                        stripped_name = "X"
                    else:
                        assert rname.startswith(provider_name_)
                        stripped_name = rname[len(provider_name_):]

                    module_name = f"v1_{kind}_{stripped_name}"
                    imports_block.line(f"from . import {module_name}")
                    module_builder = Builder()

                    module_builder.lines(_COMMON_IMPORTS)
                    module_builder.blanks(1)

                    ns_props.setdefault("model", {})[stripped_name] = f"_c.PlainDict[str, {module_name}.model]"
                    ns_props.setdefault("state", {})[stripped_name] = \
                        f"_c.ConstDict[str, _c.ConstList[{module_name}.state]]"

                    ns_paths.setdefault("model", {})[stripped_name] = ["tf", KIND_TO_KEY[kind], rname]
                    ns_paths.setdefault("state", {})[stripped_name] = [STATE_KIND_TO_KEY[kind], rname]

                    make_schema_class(
                        builder=module_builder,
                        class_name="_model",
                        schemas=[_RES_META_ARGS_SCHEMA, rschema],
                        class_path=[],
                        reader=False,
                    )

                    make_schema_class(
                        builder=module_builder,
                        class_name="_state",
                        schemas=[rschema],
                        class_path=[],
                        reader=True,
                    )

                    for what in ["model", "state"]:
                        module_builder.line(f"{what} = _{what}")

                    finalize_builder(module_name, module_builder)

                make_ns_class(
                    builder=builder,
                    class_name=f"v1_model_{kind}s",
                    props=ns_props.get("model", {}),
                    data_paths=ns_paths.get("model", {}),
                )
                make_ns_class(
                    builder=builder,
                    class_name=f"v1_state_{kind}s",
                    props=ns_props.get("state", {}),
                    data_paths=ns_paths.get("state", {}),
                )

        make_v1()

        imports_block.blanks(1)

        finalize_builder("__init__", builder)

    return _pcache.get(key, produce)


def gen_yapytfgen(
    *,
    module_dir: pathlib.Path,
    providers_paths: Mapping[str, pathlib.Path],
) -> None:
    module_fname = module_dir.joinpath("__init__.py")
    builder = Builder()

    builder.lines([
        f"from . import {provider_name} as _{provider_name}"
        for provider_name in providers_paths
    ])
    builder.lines(_COMMON_IMPORTS)
    builder.blanks(1)

    def ns(
        class_name: str,
        props: Mapping[str, str],
        data_paths: Mapping[str, List[str]] = {},
    ) -> None:
        make_ns_class(
            builder=builder,
            class_name=class_name,
            props=props,
            data_paths=data_paths,
        )

    for kind in ["data_source", "resource"]:
        ns(
            f"model_tf_v1_{kind}s",
            {
                provider_name: f"_{provider_name}.v1_model_{kind}s"
                for provider_name in providers_paths
            }
        )

    ns(
        "model_tf_v1_providers",
        {
            provider_name: f"_genbase._DictWithDefault[str, _{provider_name}.v1_model_provider]"
            for provider_name in providers_paths
        }
    )

    ns(
        "providers_model",
        {"v1": "model_tf_v1_providers"}
    )

    for kind in ["data_source", "resource"]:
        ns(
            f"state_v1_{kind}s",
            {
                provider_name: f"_{provider_name}.v1_state_{kind}s"
                for provider_name in providers_paths
            }
        )

    ns(
        "state_v1",
        {
            "d": "state_v1_data_sources",
            "r": "state_v1_resources",
        }
    )

    ns(
        "state",
        {"v1": "state_v1"}
    )

    ns(
        "model_tf_v1",
        {
            "l": "_genbase.Locals",
            "d": "model_tf_v1_data_sources",
            "r": "model_tf_v1_resources",
        }
    )

    ns(
        "model_tf",
        {"v1": "model_tf_v1"}
    )

    ns(
        "model",
        {
            "tf": "model_tf",
            "x": "_genbase.Extras",
        },
        {
            "x": ["extras"],
        }
    )

    produced = builder.produce()
    module_fname.write_text(produced)
    ast.parse(produced, filename=str(module_fname))

    for provider_name, provider_path in providers_paths.items():
        module_dir.joinpath(provider_name).symlink_to(provider_path, target_is_directory=True)



import platform


def goos() -> str:
    try:
        return {
            "Darwin": "darwin",
            "Linux": "linux",
            "Windows": "windows",
        }[platform.system()]
    except KeyError:
        raise RuntimeError("Unsupported system \"{}\"".format(platform.system()))


def goarch() -> str:
    try:
        return {
            "AMD64": "amd64",
            "x86_64": "amd64",
        }[platform.machine()]
    except KeyError:
        raise RuntimeError("Unsupported architecture \"{}\"".format(platform.machine()))



from setuptools import setup, find_packages

setup(
    name="yapytf",
    description="Yet Another Python Terraform Wrapper",
    long_description=open("README.md").read(),  # no "with..." will do for setup.py
    long_description_content_type="text/markdown; charset=UTF-8; variant=GFM",
    license="MIT",
    author="Kyrylo Shpytsya",
    author_email="kshpitsa@gmail.com",
    url="https://github.com/kshpytsya/yapytf",
    install_requires=[
        "click>=7.1.1,<8",
        "click-log>=0.3.2,<1",
        "dataclasses>=0.6,<1; python_version < '3.7'",
        "implements>=0.1.4,<1",
        # "jinja2>=2.11.1,<3",
        "jsonschema>=3.2.0,<4",
        "mypy>=0.770,<1",
        # "pytypes @ git+ssh://git@github.com/Stewori/pytypes1449eff#egg=pytypes",
        "PyYAML>=5.3.1,<6",
        "requests>=2.23.0,<3",
        "toposort>=1.5,<2",
        "tqdm>=4.45.0,<5",
        "typeguard>=2.7.1,<3",
        "typing-inspect>=0.5.0,<1",
        "xdgappdirs>=1.4.5,<2",
    ],
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    python_requires=">=3.6, <3.9",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    zip_safe=False,
    entry_points={
        "console_scripts": ["yapytf = yapytf._cli:main"]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: BSD :: FreeBSD",
        "Operating System :: POSIX :: BSD :: OpenBSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: System :: Systems Administration",
    ],
)



import textwrap

import jedi
import pytest

import yapytf._containers as cnt


prefix_code = textwrap.dedent("""
    import yapytf
    import yapytf._containers as cnt

    """)


@pytest.mark.parametrize(
    'code,ref',
    [
        (
            """
            x = cnt.DirectMutableListAccessor[int]([])
            x.
            """,
            dir(cnt.DirectMutableListAccessor[int]([]))
        ),
        (
            """
            class O(cnt.DirectMutableObjectAccessor):
                a: list
                b: str

            x = O({})
            x.
            """,
            ["a", "b"]
        ),
        (
            """
            class O(cnt.DirectMutableObjectAccessor):
                a: list
                b: str

            x = O({})
            x.a.
            """,
            dir(list)
        ),
        (
            """
            class O(cnt.DirectMutableObjectAccessor["O"]):
                a: list

            x = O({})
            with x as y:
                y.a.
            """,
            dir(list)
        ),
    ]
)
def test(code: str, ref: str) -> None:
    code = prefix_code + textwrap.dedent(code)
    lines = code.splitlines()
    script = jedi.Script(
        source=code,
        line=len(lines),
        column=len(lines[-1]),
        path="source.py",
    )

    def nounders(l):
        return (i for i in l if not i.startswith("_"))

    completions = set(nounders(i.name for i in script.completions()))
    expected_completions = set(nounders(ref))
    assert completions == expected_completions



import logging
import os
import pathlib
import stat
import urllib.parse
import zipfile

import click
import requests
import tqdm

from . import _pcache


logger = logging.getLogger(__name__)


# Per https://stackoverflow.com/questions/42326428/zipfile-in-python-file-permission
def _extract_all_with_executable_permission(zf, target_dir):
    ZIP_UNIX_SYSTEM = 3

    for info in zf.infolist():
        extracted_path = zf.extract(info, target_dir)

        if info.create_system == ZIP_UNIX_SYSTEM and os.path.isfile(extracted_path):
            unix_attributes = info.external_attr >> 16
            if unix_attributes & stat.S_IXUSR:
                os.chmod(extracted_path, os.stat(extracted_path).st_mode | stat.S_IXUSR)


def _download(url: str, dest: pathlib.Path, *, what: str = None):
    logger.debug("downloading %s", url)
    req = requests.get(url, stream=True)
    try:
        req.raise_for_status()
    except requests.HTTPError as e:
        raise click.ClickException(str(e))

    total_size = int(req.headers.get("content-length", 0))
    block_size = 1024
    written = 0

    with dest.open("wb") as f, tqdm.tqdm(
        desc="Downloading {}".format(what or url),
        total=total_size,
        unit="B",
        unit_divisor=1024,
        unit_scale=True,
    ) as pbar:
        for data in req.iter_content(block_size):
            f.write(data)
            pbar.update(len(data))
            written = written + len(data)

    if total_size and written != total_size:
        raise RuntimeError(
            "Download size mismatch: expected {} got {} bytes".format(
                total_size, written
            )
        )


def get(url: str, *, unzip: bool = True, keep_zip: bool = False) -> pathlib.Path:
    def produce(dir_path: pathlib.Path) -> None:
        url_fname = pathlib.PurePosixPath(urllib.parse.urlparse(url).path).name
        download_path = dir_path.joinpath(url_fname)

        _download(url, download_path, what=url_fname)

        if unzip and zipfile.is_zipfile(download_path):
            with zipfile.ZipFile(download_path) as zf:
                _extract_all_with_executable_permission(zf, dir_path)

            if not keep_zip:
                download_path.unlink()

    return _pcache.get(url, produce)
